--[[
Name: targetAssist
Author: Sonora (The Dragonflight, US Kirin Tor)
Website & SVN: http://wow.curseforge.com/addons/targetassist/
License: GNU General Public License v3
]]

--Utility functions
local function keyFromValue(t, value)
    for k,v in pairs(t) do
        if v == value then return k end
    end
end

local function keys(t)
    ks = {}
    for k,_ in pairs(t) do table.insert(ks, k) end
    return ks
end

local function hasValue(t, value)
    for _, v in ipairs(t) do
        if v == value then return true end
    end
    for k, v in pairs(t) do
        if v == value then return true end
    end
end

local function tableIndex(t, value)
    for idx, v in ipairs(t) do
        if v == value then return idx end
    end
end

local function inverted(t)

    local newTable = {}
    for i = 0,#t-1 do
        table.insert(newTable, t[#t-i])
    end
    return newTable

end

--Methods that draw stuff
function targetAssist:yesnoBox(msg, callback)
    StaticPopupDialogs["unitDoodles_YESNOBOX"] = {
        text = msg,
        button1 = "Yes",
        button2 = "No",
        OnAccept = function(self, data, data2)
            callback()
        end,
        timeout = 0,
        whileDead = true,
        hideOnEscape = true,
    }
    
    StaticPopup_Show ("unitDoodles_YESNOBOX")
end

function targetAssist:showHeaderMenu()

    self:updateRoster()

    local groupSelection = {}
    local unitNames = keys(self.roster)
    table.sort(unitNames)
    for _, unitName in ipairs(unitNames) do
        table.insert(groupSelection, {
            text = unitName,
            checked = function() return hasValue(self.config.targets, unitName) end,
            func = function(this, arg1, arg2, checked)
                if checked then
                    table.remove(self.config.targets, tableIndex(self.config.targets, unitName))
                    this.checked = false
                else
                    table.insert(self.config.targets, unitName)
                    this.checked = true
                end
                self:updateConfig()
            end,
        })
    end
    
    local broadcastSelection = {}
    for _, broadcast in ipairs(self.previousBroadcasts) do
        local sender, dist, targetList = unpack(broadcast)
        table.insert(broadcastSelection, {
                text = sender..' ('..table.concat(targetList, ", ")..')',
                notCheckable = true,
                func = function()
                    self.config.targets = targetList
                    self:updateConfig()
                end,
            }
        )
    end

    self.headerMenu = {
        {text = 'Main assists', 
            isTitle = true,
            notCheckable = true,
        },
    
        {text = 'Select custom assists',
            hasArrow = true,
            notCheckable = true,
            menuList = groupSelection,
        },
        {text = 'Select all',
            notCheckable = true,
            func = function() self:addAll() end,
        },
        {text = 'Clear all',
            notCheckable = true,
            func = function() self:clearAll() end,
        },
        {text = 'Add party',
            notCheckable = true,
            func = function() self:addParty() end,
        },        
        {text = 'Add all tanks',
            notCheckable = true,
            func = function()
                self:addTanks()
                self:updateConfig()
            end,
        },
        {text = 'Add all main assists',
            notCheckable = true,
            func = function()
                self:addMainAssists()
                self:updateConfig()
            end,
        },
        {text = 'Add current target',
            notCheckable = true,
            func = function() 
                self:addTarget()
                self:updateConfig()
            end,
        },
        {text = 'Invert assist order',
            notCheckable = true,
            func = function() 
                self:invertTargetOrder()
            end,
        },
    
        {text = 'Broadcasts', 
            isTitle = true,
            notCheckable = true,
        },
        {text = 'Recieved broadcasts',
            hasArrow = true,
            notCheckable = true,
            menuList = broadcastSelection,
        },
        {text = 'Broadcast targets',
            hasArrow = true,
            notCheckable = true,
            menuList = {
                {text = 'To raid',
                    notCheckable = true,
                    func = function() 
                        self:SendCommMessage('targetAssist.broadcastTargets', self:Serialize(self.config.targets), 'RAID')
                    end,
                },
                {text = 'To party',
                    notCheckable = true,
                    func = function() 
                        self:SendCommMessage('targetAssist.broadcastTargets', self:Serialize(self.config.targets), 'PARTY')
                    end,
                },
                {text = 'To guild',
                    notCheckable = true,
                    func = function() 
                        self:SendCommMessage('targetAssist.broadcastTargets', self:Serialize(self.config.targets), 'GUILD')
                    end,
                },
            },
        },
    
        {text = 'Configuration',
            isTitle = true,
            notCheckable = true,
        },
        {text = 'Open options panel',
            notCheckable = true,
            func = function() InterfaceOptionsFrame_OpenToCategory(self.optionsBaseFrame) end,
        },
    }

    -- Make the menu appear at the cursor: 
    EasyMenu(self.headerMenu, self.contextMenu, "cursor", 0 , 0, "MENU")
    
end

function targetAssist:showAssistFrameMenu(targetName)

    local idx = tableIndex(self.config.targets, targetName)
    local assistMenu = {
        {text = 'Move up',
            notCheckable = true,
            disabled = idx==1, 
            func = function()
                table.remove(self.config.targets, idx)
                table.insert(self.config.targets, idx-1, targetName)
                self:updateConfig()
            end,
        },
        {text = 'Move down',
            notCheckable = true,
            disabled = idx == #self.config.targets,
            func = function()
                table.remove(self.config.targets, idx)
                table.insert(self.config.targets, idx+1, targetName)
                self:updateConfig()
            end,
        },
    }
    
    EasyMenu(assistMenu, self.contextMenu, "cursor", 0 , 0, "MENU")
end

function targetAssist:showAssistTargetMenu(unitID)

    local currentIcon = GetRaidTargetIndex(unitID)

    local menu = {}
    for iconIdx, raidIcon in ipairs(self.config.raidIcons) do
        local iconName, iconTexture = unpack(raidIcon)
        table.insert(menu, {
                text = iconName,
                checked = iconIdx == currentIcon,
                func = function(this, arg1, arg2, checked)
                    if checked == true then
                        this.checked = true
                    else
                        this.checked = false
                    end
                    SetRaidTargetIcon(unitID, iconIdx)
                end,
                icon = iconTexture,
            }
        )
    
    end

    EasyMenu(menu, self.contextMenu, "cursor", 0 , 0, "MENU")

end

function targetAssist:createHeaderFrame()

    if not self.headerFrame then
        self.headerFrame = CreateFrame('Frame', 'targetAssistHeaderFrame', UIParent)
    end
    
    self.headerFrame:ClearAllPoints()
    self.headerFrame:SetWidth(self.config.headerWidth)
    self.headerFrame:SetHeight(self.config.headerHeight)
    self.headerFrame:SetPoint(unpack(self.config.headerAnchor))
    
    self.headerFrame:SetBackdrop(self.config.headerBackdrop)
    self.headerFrame:SetBackdropColor(0.15, 0.15, 0.15)
    
    if not self.headerFrame.texture then
        self.headerFrame.texture = self.headerFrame:CreateTexture()
    end
    self.headerFrame.texture:SetAllPoints(self.headerFrame)
    self.headerFrame.texture:SetTexture(self.config.headerTexture)
    self.headerFrame.texture:SetVertexColor(unpack(self.config.headerColor))
    
    self.headerFrame.text = self.headerFrame:CreateFontString(nil, 'OVERLAY')
    self.headerFrame.text:SetFont(self.config.fontName, self.config.fontHeight)
    self.headerFrame.text:SetTextColor(unpack(self.config.fontColor))
    self.headerFrame.text:SetAllPoints(self.headerFrame)
    self.headerFrame.text:SetText("targetAssist")
    if self.config.hideAddonName then self.headerFrame.text:Hide() end
    
    self.headerFrame:EnableMouse(true)
    self.headerFrame:SetMovable(true)
    self.headerFrame:RegisterForDrag('LeftButton')
    self.headerFrame:SetScript("OnDragStart", function() 
        if self.config.headerUnlocked then self.headerFrame:StartMoving() end
    end)
    
    self.headerFrame:SetScript("OnDragStop", 
        function()
            self.headerFrame:StopMovingOrSizing()
            self.config.headerAnchor = {self.headerFrame:GetPoint()}
            self.config.headerAnchor[2] = 'UIParent'
        end
    )
    
    self.headerFrame:SetScript("OnMouseUp",
        function(_, button)
            if button == 'RightButton' then
                self:showHeaderMenu()
            end
        end
    )
    
    self.headerFrame:Show()
    
end

function targetAssist:createTargetButtons()
    
    local headerFrame = self.headerFrame
    local targetTable = self.config.targets
    if self.config.invertTargetOrder then
        targetTable = inverted(self.config.targets)
    end
    
    --Support for Clique
    ClickCastFrames = ClickCastFrames or {}

    --Clean out target button structures (TODO:  how expensive are GC cycles for WoW lua?  I tend to hit them hard...)
    if self.targetButtons then
        for targetName, buttonFrame in pairs(self.targetButtons) do
            buttonFrame:Hide()
            buttonFrame:SetParent(nil)
            ClickCastFrames[buttonFrame] = nil
        end
    end
    self.targetButtons = {}
    
    --For each assist target
    local anchorFrame, width = headerFrame, self.config.headerWidth/2
    self.buttonWidth = width-self.config.spacingOffset
    for i, targetName in ipairs(targetTable) do
    
        local unitID = self.roster[targetName]
        if unitID then
    
            --Make the assist target unit button
            local assistFrame = CreateFrame("Button", 'targetAssistAssistFrame'..i, UIParent, "SecureUnitButtonTemplate")
            assistFrame:SetAttribute('unitName',targetName)
        
            local buttonAnchorPoint, headerAnchorPoint, verticalOffset = unpack(self.config.targetButtonAnchor)
            assistFrame:SetPoint(buttonAnchorPoint, anchorFrame, headerAnchorPoint, 0, verticalOffset)
            assistFrame:SetWidth(self.buttonWidth)
            assistFrame:SetHeight(self.config.targetHeight)
            
            assistFrame:SetBackdrop(self.config.targetBackdrop)
            assistFrame:SetBackdropColor(0.15, 0.15, 0.15)
            
            local frameTexture = assistFrame:CreateTexture()
            frameTexture:SetPoint('TOP', assistFrame, 'TOP')
            frameTexture:SetPoint('BOTTOM', assistFrame, 'BOTTOM')
            frameTexture:SetPoint('LEFT', assistFrame, 'LEFT')
            frameTexture:SetWidth(self.buttonWidth)
            frameTexture:SetTexture(self.config.targetTexture)
            frameTexture:SetVertexColor(unpack(self.config.friendlyTargetColor))
            assistFrame.texture = frameTexture
        
            assistFrame.fontString = assistFrame:CreateFontString(nil, 'OVERLAY')
            assistFrame.fontString:SetFont(self.config.fontName, self.config.fontHeight)
            assistFrame.fontString:SetTextColor(unpack(self.config.fontColor))
            assistFrame.fontString:SetAllPoints(assistFrame)
        
            assistFrame.mostTargetedIcon = assistFrame:CreateTexture(nil, 'OVERLAY')
            assistFrame.mostTargetedIcon:SetPoint('BOTTOMRIGHT', assistFrame, 'BOTTOMLEFT', -2, 0)
            assistFrame.mostTargetedIcon:SetHeight(self.config.targetHeight/2)
            assistFrame.mostTargetedIcon:SetWidth(self.config.targetHeight/2)
            assistFrame.mostTargetedIcon:Hide()
        
            assistFrame.targetCountString = assistFrame:CreateFontString(nil, 'OVERLAY')
            assistFrame.targetCountString:SetFont(self.config.fontName, self.config.fontHeight)
            assistFrame.targetCountString:SetTextColor(unpack(self.config.fontColor))
            assistFrame.targetCountString:SetPoint('TOPRIGHT', assistFrame, 'TOPLEFT', -2, 0)
            assistFrame.targetCountString:SetHeight(self.config.targetHeight/2)
            assistFrame.targetCountString:SetWidth(self.config.targetHeight/2)
        
            assistFrame:SetScript("OnMouseUp",
                function(_, button)
                    if button == 'RightButton' then
                        if not IsModifierKeyDown() then self:showAssistFrameMenu(targetName) end
                    end
                end
            )
        
            assistFrame:SetAttribute('unitID', unitID)
            assistFrame.unitID = unitID
            SecureUnitButton_OnLoad(assistFrame, unitID, 'print("menu")')
        
            ClickCastFrames[assistFrame] = true
        
            assistFrame:Show()
            self.targetButtons[targetName] = assistFrame
        
            --Make the assist target target unit button
            local targetFrame = CreateFrame("Button", 'targetAssistTargetFrame'..i, UIParent, "SecureUnitButtonTemplate")
        
            targetFrame:SetPoint('TOP', assistFrame, 'TOP')
            targetFrame:SetPoint('RIGHT', headerFrame, 'RIGHT')
            targetFrame:SetWidth(self.buttonWidth)
            targetFrame:SetHeight(self.config.targetHeight)
            
            targetFrame:SetBackdrop(self.config.targetBackdrop)
            targetFrame:SetBackdropColor(0.15, 0.15, 0.15)
            
            local frameTexture = targetFrame:CreateTexture()
            frameTexture:SetPoint('TOP', targetFrame, 'TOP')
            frameTexture:SetPoint('BOTTOM', targetFrame, 'BOTTOM')
            frameTexture:SetPoint('LEFT', targetFrame, 'LEFT')
            frameTexture:SetWidth(self.buttonWidth)
            frameTexture:SetTexture(self.config.targetTexture)
            frameTexture:SetVertexColor(unpack(self.config.friendlyTargetColor))
            targetFrame.texture = frameTexture

            targetFrame.fontString = targetFrame:CreateFontString(nil, 'OVERLAY')
            targetFrame.fontString:SetFont(self.config.fontName, self.config.fontHeight)
            targetFrame.fontString:SetTextColor(unpack(self.config.fontColor))
            targetFrame.fontString:SetAllPoints(targetFrame)
        
            targetFrame.mostTargetedIcon = targetFrame:CreateTexture(nil, 'OVERLAY')
            targetFrame.mostTargetedIcon:SetPoint('BOTTOMLEFT', targetFrame, 'BOTTOMRIGHT', 2, 0)
            targetFrame.mostTargetedIcon:SetHeight(self.config.targetHeight/2)
            targetFrame.mostTargetedIcon:SetWidth(self.config.targetHeight/2)
            targetFrame.mostTargetedIcon:Hide()
        
            targetFrame.targetCountString = targetFrame:CreateFontString(nil, 'OVERLAY')
            targetFrame.targetCountString:SetFont(self.config.fontName, self.config.fontHeight)
            targetFrame.targetCountString:SetTextColor(unpack(self.config.fontColor))
            targetFrame.targetCountString:SetPoint('TOPLEFT', targetFrame, 'TOPRIGHT', 2, 0)
            targetFrame.targetCountString:SetHeight(self.config.targetHeight/2)
            targetFrame.targetCountString:SetWidth(self.config.targetHeight/2)
        
            targetFrame.raidIcon = targetFrame:CreateTexture(nil, "OVERLAY", nil)
            targetFrame.raidIcon:SetPoint('BOTTOMRIGHT', targetFrame, 'BOTTOMRIGHT', 0, 0)
            targetFrame.raidIcon:SetHeight(self.config.targetHeight*.75)
            targetFrame.raidIcon:SetWidth(self.config.targetHeight*.75)
        
            targetFrame:SetScript("OnMouseUp",
                function(_, button)
                    if button == 'RightButton' then
                        if not IsModifierKeyDown() then self:showAssistTargetMenu(unitID..'target') end
                    end
                end
            )
        
            targetFrame:SetAttribute('unitID', unitID..'target')
            targetFrame.unitID = unitID..'target'
            SecureUnitButton_OnLoad(targetFrame, unitID..'target', 'print("menu")')
        
            ClickCastFrames[targetFrame] = true
        
            targetFrame:Show()        
            self.targetButtons[targetName..'target'] = targetFrame
            
            --Optionally make a target of target button
            if self.config.showTargetTarget then
                local targetTargetFrame = CreateFrame("Button", nil, UIParent, "SecureUnitButtonTemplate")
            
                targetTargetFrame:SetPoint('TOP', assistFrame, 'TOP')
                targetTargetFrame:SetPoint('LEFT', headerFrame, 'RIGHT', (self.config.targetHeight/2)+5, 0)
                targetTargetFrame:SetWidth(self.buttonWidth)
                targetTargetFrame:SetHeight(self.config.targetHeight)
                
                local frameTexture = targetTargetFrame:CreateTexture()
                frameTexture:SetPoint('TOP', targetTargetFrame, 'TOP')
                frameTexture:SetPoint('BOTTOM', targetTargetFrame, 'BOTTOM')
                frameTexture:SetPoint('LEFT', targetTargetFrame, 'LEFT')
                frameTexture:SetWidth(self.buttonWidth)
                frameTexture:SetTexture(self.config.targetTexture)
                frameTexture:SetVertexColor(unpack(self.config.friendlyTargetColor))
                targetTargetFrame.texture = frameTexture

                targetTargetFrame.fontString = targetTargetFrame:CreateFontString(nil, 'OVERLAY')
                targetTargetFrame.fontString:SetFont(self.config.fontName, self.config.fontHeight)
                targetTargetFrame.fontString:SetTextColor(unpack(self.config.fontColor))
                targetTargetFrame.fontString:SetAllPoints(targetTargetFrame)
                
                targetTargetFrame:SetAttribute('unitID', unitID..'targettarget')
                targetTargetFrame.unitID = unitID..'targettarget'
                SecureUnitButton_OnLoad(targetTargetFrame, unitID..'targettarget', 'print("menu")')
            
                ClickCastFrames[targetTargetFrame] = true
            
                targetFrame:Show()        
                self.targetButtons[targetName..'targettarget'] = targetTargetFrame
            
            end
            
            --Next time around, anchor to this row
            anchorFrame = assistFrame
    
        end
    
    end
    
end

function targetAssist:createBasePanel()

    return {
        type = "group",
        name = "targetAssist Configuration",
        args={
            introduction = {
                type = 'description',
                name = "Changes to many of the configuration settings for targetAssist can only be updated when you are out of combat.  If you make changes while in combat, they will be queued and applied later.\n\nIf you'd like to change the texture or font settings please make sure you have |cffFF7D0ASharedMedia|r installed.",
                order = 1
            },
        
            h1 = {
                type = 'header',
                name = 'Button Header',
                order = 20,
            },
            headerUnlocked = {
                type = 'toggle',
                name = 'Unlock header',
                desc = 'Allows the header bar to be moved by dragging it with the mouse.  Uncheck to lock it in place.',
                width = 'full',
                get = function() return self.config.headerUnlocked end,
                set = function(info, value) self.config.headerUnlocked = value end,
                order = 21,
            },
            hideOutOfGroup = {
                type = 'toggle',
                name = 'Hide addon when not in a group',
                desc = 'Hides the header bar and any targeting buttons when you are not in a party/raid.  Uncheck to leave the header bar shown at all times.',
                width = 'full',
                get = function() return self.config.hideOutOfGroup end,
                set = function(info, value) 
                    self.config.hideOutOfGroup = value 
                    self:updateRoster()
                end,
                order = 22,
            },
            hideAddonName = {
                type = 'toggle',
                name = 'Hide addon name',
                desc = 'Removes the text "targetAssist" from the header bar.',
                width = 'full',
                get = function() return self.config.hideAddonName end,
                set = function(info, value) 
                    self.config.hideAddonName = value 
                    if value then
                        self.headerFrame.text:Hide()
                    else
                        self.headerFrame.text:Show()
                    end
                end,
                order = 22.5,
            },
            headerWidth = {    
                type = 'input',
                name = 'Header width',
                desc = 'Width of the header bar (a positive value)',
                width = 'half',
                pattern = '%d',
                get = function() return tostring(self.config.headerWidth) end,
                set = function(info, value) 
                    self.config.headerWidth = tonumber(value)
                    self:updateConfig(true)
                end,
                order = 23,
            },
            headerHeight = {    
                type = 'input',
                name = 'Header height',
                desc = 'Height of the header bar (a positive value)',
                width = 'half',
                pattern = '%d',
                get = function() return tostring(self.config.headerHeight) end,
                set = function(info, value) 
                    self.config.headerHeight = tonumber(value)
                    self:updateConfig(true)
                end,
                order = 24,
            },
            headerTexture = {
                type = 'select',
                name = 'Header texture',
                desc = 'Texture used to paint the header bar.  If this control is greyed out please installed SharedMedia.',
                dialogControl = 'LSM30_Statusbar',
                values = AceGUIWidgetLSMlists.statusbar,
                get = function() return keyFromValue(AceGUIWidgetLSMlists.statusbar, self.config.headerTexture) end,
                set = function(info, key)
                    self.config.headerTexture = AceGUIWidgetLSMlists.statusbar[key]
                    self:updateConfig(true)
                end,
                order = 25,
            },
            headerColor = {
                type = 'color',
                name = 'Header color',
                desc = 'Changes the color of the header bar.',
                hasAlpha = true,
                get = function() return unpack(self.config.headerColor) end,
                set = function(info, r,g,b,a) 
                    self.config.headerColor = {r,g,b,a} 
                    self:updateConfig(true)
                end,
                order = 26,
            },
        
            h2 = {
                type = 'header',
                name = 'Assist Buttons',
                order = 30,
            },
            targetHeight = {    
                type = 'input',
                name = 'Button height',
                desc = 'Height of the assist buttons (a positive value)',
                width = 'half',
                pattern = '%d',
                get = function() return tostring(self.config.targetHeight) end,
                set = function(info, value) 
                    self.config.targetHeight = tonumber(value)
                    self:updateConfig(true)
                end,
                order = 32,
            },
            targetTexture = {
                type = 'select',
                name = 'Button texture',
                desc = 'Texture used to paint the assist buttons.  If this control is greyed out please installed SharedMedia.',
                dialogControl = 'LSM30_Statusbar',
                values = AceGUIWidgetLSMlists.statusbar,
                get = function() return keyFromValue(AceGUIWidgetLSMlists.statusbar, self.config.targetTexture) end,
                set = function(info, key)
                    self.config.targetTexture = AceGUIWidgetLSMlists.statusbar[key]
                    self:updateConfig(true)
                end,
                order = 34,
            },
            friendlyTargetColor = {
                type = 'color',
                name = 'Friendly target color',
                desc = 'Changes the color of the assist buttons tracking friendly targets.',
                hasAlpha = true,
                get = function() return unpack(self.config.friendlyTargetColor) end,
                set = function(info, r,g,b,a) 
                    self.config.friendlyTargetColor = {r,g,b,a} 
                    self:updateConfig(true)
                end,
                order = 35,
            },
            hostileTargetColor = {
                type = 'color',
                name = 'Hostile target color',
                desc = 'Changes the color of the assist buttons tracking hostile targets.',
                hasAlpha = true,
                get = function() return unpack(self.config.hostileTargetColor) end,
                set = function(info, r,g,b,a) 
                    self.config.hostileTargetColor = {r,g,b,a} 
                    self:updateConfig(true)
                end,
                order = 35,
            },
            missingColor = {
                type = 'color',
                name = 'Button color (dead or missing)',
                desc = 'Color for buttons when the associated unit is dead or missing.',
                hasAlpha = true,
                width = 'full',
                get = function() return unpack(self.config.missingColor) end,
                set = function(info, r,g,b,a) 
                    self.config.missingColor = {r,g,b,a} 
                    self:updateConfig(true)
                end,
                order = 36,
            },
            myTargetColor = {
                type = 'color',
                name = 'Button color (my target)',
                desc = 'Color for buttons associated with the unit you currently have targeted.',
                hasAlpha = true,
                width = 'full',
                get = function() return unpack(self.config.myTargetColor) end,
                set = function(info, r,g,b,a) 
                    self.config.myTargetColor = {r,g,b,a} 
                    self:updateConfig(true)
                end,
                order = 37,
            },
        
            showHealth = {
                type = 'toggle',
                name = 'Show health',
                desc = 'If checked the colored texture on assist buttons will reflect the percentage health of the associated unit.',
                width = 'full',
                get = function() return self.config.showHealth end,
                set = function(info, value) 
                    self.config.showHealth = value
                    if value == false then
                        for _, buttonFrame in pairs(self.targetButtons or {}) do
                            buttonFrame.texture:SetWidth( self.buttonWidth )
                        end
                    end
                end,
                order = 38,
            },
        
            h3 = {
                type = 'header',
                name = 'Fonts & Style',
                order = 40,
            },
        
            fontName = {
                type = 'select',
                name = 'Font',
                desc = 'Font used to draw text in the addon.  If this box is greyed out please install SharedMedia.',
                dialogControl = 'LSM30_Font',
                values = AceGUIWidgetLSMlists.font,
                get = function() return keyFromValue(AceGUIWidgetLSMlists.font, self.config.fontName) end,
                set = function(info, key)
                    self.config.fontName = AceGUIWidgetLSMlists.font[key]
                    self:updateConfig(true)
                end,
                order = 41,
            },
            fontHeight = {    
                type = 'input',
                name = 'Font height',
                desc = 'Height of text drawn by the addon (a positive value)',
                width = 'half',
                pattern = '%d',
                get = function() return tostring(self.config.fontHeight) end,
                set = function(info, value) 
                    self.config.fontHeight = tonumber(value)
                    self:updateConfig(true)
                end,
                order = 42,
            },
            fontColor = {
                type = 'color',
                name = 'Font color',
                desc = 'Color for text drawn by the addon.',
                hasAlpha = true,
                get = function() return unpack(self.config.fontColor) end,
                set = function(info, r,g,b,a) 
                    self.config.fontColor = {r,g,b,a} 
                    self:updateConfig(true)
                end,
                order = 43,
            },
            spacingOffset = {    
                type = 'input',
                name = 'Horizontal spacing',
                desc = 'Horizontal spacing between assist buttons',
                pattern = '%d',
                width = 'half',
                get = function() return tostring(self.config.spacingOffset) end,
                set = function(info, value) 
                    self.config.spacingOffset = tonumber(value)
                    self:updateConfig(true)
                end,
                order = 45,
            },
            verticalOffset = {    
                type = 'input',
                name = 'Vertical spacing',
                desc = 'Vertical spacing between rows of assist buttons and the header bar.',
                pattern = '%d',
                width = 'half',
                get = function() return tostring(self.config.targetButtonAnchor[3]*(self.config.targetButtonAnchor[1] == 'BOTTOMLEFT' and 1 or -1)) end,
                set = function(info, value) 
                    self.config.targetButtonAnchor[3] = tonumber(value)*(self.config.targetButtonAnchor[1] == 'BOTTOMLEFT' and 1 or -1)
                    self:updateConfig()
                end,
                order = 46,
            },
            growVertically = {
                type = 'toggle',
                name = 'Advance buttons upwards',
                desc = 'If checked the targeting buttons advance upwards from the header bar.  If unchecked they advance downwards.',
                width = 'full',
                get = function() 
                    return self.config.targetButtonAnchor[1] == 'BOTTOMLEFT'
                end,
                set = function(info, value)
                    if value == true then
                        self.config.targetButtonAnchor[1] = 'BOTTOMLEFT'
                        self.config.targetButtonAnchor[2] = 'TOPLEFT'
                        self.config.invertTargetOrder = true
                    else
                        self.config.targetButtonAnchor[1] = 'TOPLEFT'
                        self.config.targetButtonAnchor[2] = 'BOTTOMLEFT'
                        self.config.invertTargetOrder = false
                    end
                    self.config.targetButtonAnchor[3] = self.config.targetButtonAnchor[3]*-1
                    self:updateConfig()
                end,
                order = 44,
            },
        
            h5 = {
                type = 'header',
                name = 'Optional Features',
                order = 10,
            },
            autoAcceptBroadcasts = {
                type = 'toggle',
                name = 'Accept all broadcasts',
                desc = 'Automatically accept incomming assist target lists broadcasted by other players in your party/group.  If this is left unchecked you will be asked whether or not you want to accept each broadcast individually.',
                width = 'full',
                get = function() return self.config.autoAcceptBroadcasts end,
                set = function(info, value) self.config.autoAcceptBroadcasts = value end,
                order = 11,
            },
            autoUpdateTanks = {
                type = 'toggle',
                name = 'Automatically add tanks',
                desc = 'Automatically add any party/raid members assigned the roles of tank or main tank.  If left unchecked you can manually add all tanks from the header bar context menu (right click).\n\nRaid leaders can set the main tank role with the "/mt PlayerName" chat command.',
                width = 'full',
                get = function() return self.config.autoUpdateTanks end,
                set = function(info, value) self.config.autoUpdateTanks = value end,
                order = 12,
            },
            autoUpdateMainAssists = {
                type = 'toggle',
                name = 'Automatically add main assists',
                desc = 'Automatically add any party/raid members assigned the role of main assist through the Blizzard UI.  If left unchecked you can manually add all main assists from the header bar context menu (right click).\n\nRaid leaders can set the main assist role with the "/ma PlayerName" chat command.',
                width = 'full',
                get = function() return self.config.autoUpdateMainAssists end,
                set = function(info, value) self.config.autoUpdateMainAssists = value end,
                order = 13,
            },
            autoUpdateORA2 = {
                type = 'toggle',
                name = 'Include tanks from oRA2',
                desc = 'Include party/raid members assigned as tanks through oRA2 when adding tanks manually or automatically (requires oRA2).',
                width = 'full',
                get = function() return self.config.autoUpdateORA2 end,
                set = function(info, value) self.config.autoUpdateORA2 = value end,
                order = 13.2,
            },
            autoUpdateORA3 = {
                type = 'toggle',
                name = 'Include tanks from oRA3',
                desc = 'nclude party/raid members assigned as tanks through oRA3 when adding tanks manually or automatically (requires oRA3).',
                width = 'full',
                get = function() return self.config.autoUpdateORA3 end,
                set = function(info, value) self.config.autoUpdateORA3 = value end,
                order = 13.4,
            },
            trackNumberTargetingHostiles = {
                type = 'toggle',
                name = 'Track group members targeting hostiles',
                desc = 'Adds a counter next to all hostile units being targeted by MTs/MAs showing the number of party/raid members targeting that unit.',
                width = 'full',
                get = function() return self.config.trackNumberTargetingHostiles end,
                set = function(info, value) 
                    self.config.trackNumberTargetingHostiles = value
                    self:updateConfig()
                end,
                order = 14,
            },
            markMostTargetedHostile = {
                type = 'toggle',
                name = 'Mark most popular hostile target',
                desc = 'Marks the most common hostile target with a Skull icon.',
                width = 'full',
                get = function() return self.config.markMostTargetedHostile end,
                set = function(info, value) 
                    self.config.markMostTargetedHostile = value
                    self:updateConfig()
                end,
                order = 15,
            },
            trackNumberTargetingFriendlies = {
                type = 'toggle',
                name = 'Track group members targeting friendlies',
                desc = 'Adds a counter next to all MAs/MTs showing the number of party/raid members targeting that player.',
                width = 'full',
                get = function() return self.config.trackNumberTargetingFriendlies end,
                set = function(info, value) 
                    self.config.trackNumberTargetingFriendlies = value
                    self:updateConfig()
                end,
                order = 16,
            },
            markMostTargetedFriendly = {
                type = 'toggle',
                name = 'Mark most popular friendly target',
                desc = 'Marks the most common target with a green triangle icon.',
                width = 'full',
                get = function() return self.config.markMostTargetedFriendly end,
                set = function(info, value) 
                    self.config.markMostTargetedFriendly = value
                    self:updateConfig()
                end,
                order = 17,
            },
            fadeOutOfRange = {
                type = 'toggle',
                name = 'Fade out of range units',
                desc = 'Fades unit buttons for targets that are out of range.  Friendly units are faded based on the range of typical heals.  Hostile targets are faded based on class-specific spell/ability checks.',
                width = 'full',
                get = function() return self.config.fadeOutOfRange end,
                set = function(info, value) self.config.fadeOutOfRange = value end,
                order = 18,
            },
            showTargetTarget = {
                type = 'toggle',
                name = 'Show target of target',
                desc = "Adds an extra column of unit buttons showing each assist target's target",
                width = 'full',
                get = function() return self.config.showTargetTarget end,
                set = function(info, value) 
                    self.config.showTargetTarget = value 
                    self:updateConfig()
                end,
                order = 19,
            },
            includePlayer = {
                type = 'toggle',
                name = 'Include me in the roster',
                desc = "Allows you to add yourself as a main assist.  If unchecked you will not show up in the selection list and will not be show as an assist button.",
                width = 'full',
                get = function() return self.config.includePlayer end,
                set = function(info, value) 
                    self.config.includePlayer = value 
                    self:updateRoster()
                    self:updateConfig()
                end,
                order = 19.25,
            },
            showPets = {
                type = 'toggle',
                name = 'Show pets',
                desc = "Includes your pet and all group pets in the roster.",
                width = 'full',
                get = function() return self.config.showPets end,
                set = function(info, value)
                    self.config.showPets = value
                    self:updateRoster()
                    self:updateConfig()
                end,
                order = 19.5,
            },
        
            h6 = {
                type = 'header',
                name = 'Performance',
                order = 60,
            },
            t1 = {
                type = 'description',
                name = 'It is recommended that you reload your user interface after making adjustments to the configuration.  This will help to free up memory and make the game more responsive.',
                order = 61,
            },
            reloadUI = {
                type = 'execute',
                name = 'Reload',
                desc = 'Reloads the WoW user interface.',
                func = function() ReloadUI() end,
                order = 62,
            },
            updateInterval = {    
                type = 'input',
                name = 'Update interval',
                desc = 'Time (in seconds) between updates of button state. Change this to a larger value if you are having performance problems (recommended values are 0.1-0.5).',
                width = 'half',
                pattern = '%d',
                get = function() return tostring(self.config.updateInterval) end,
                set = function(info, value) 
                    self.config.updateInterval = tonumber(value)
                    self:updateConfig(true)
                end,
                order = 63,
            },
            broadcastHistory = {
                type = 'input',
                name = 'Broadcast history',
                desc = 'Number of past broadcasts to save in the broadcast history.',
                pattern = '%d',
                get = function() return tostring(self.config.broadcastHistory) end,
                set = function(info, value) 
                    self.config.broadcastHistory = tonumber(value)
                    while #self.previousBroadcasts > self.config.broadcastHistory do
                        table.remove(self.broadcastHistory, 1)
                    end
                    self:updateConfig(true)
                end,
                order = 64,
            },
            disabled = {
                type = 'toggle',
                name = 'Disable addon',
                desc = "If checked the addon will be hidden and updates disabled.",
                width = 'full',
                get = function() return self.disabled end,
                set = function(info, value) self.disabled = value end,
                order = 65,
            },
        
        }
    }

end

function targetAssist:createConsoleOptions()

    return {
        type = "group",
        name = "targetAssist Configuration",
        args={
            introduction = {
                type = 'description',
                name = "Changes to many of the configuration settings for targetAssist can only be updated when you are out of combat.  If you make changes while in combat, they will be queued and applied later.",
                order = 1
            },
            add = {
                type = 'group',
                name = 'Add MAs/MTs',
                order = 20,
                args = {
                    target = {
                        type = 'execute',
                        name = 'add current target',
                        order = 1,
                        func = function() self:addTarget() end,
                    },
                    all = {
                        type = 'execute',
                        name = 'add all group members',
                        order = 2,
                        func = function() self:addAll() end,
                    },
                    party = {
                        type = 'execute',
                        name = 'add all party members',
                        order = 3,
                        func = function() self:addParty() end,
                    },
                    tanks = {
                        type = 'execute',
                        name = 'add all tanks',
                        order = 4,
                        func = function() self:addTanks() end,
                    },
                    mainassists = {
                        type = 'execute',
                        name = 'add all main assists',
                        order = 4,
                        func = function() self:addMainAssists() end,
                    },
                    player = {
                        type = 'input',
                        name = 'Add player by name (must be in raid or party).  Usage: "/ta add player NAME" where NAME is the name of the player to add.',
                        order = 4,
                        get = function() return '' end,
                        set = function(info, value)
                            if self.roster[value] then
                                if not hasValue(self.config.targets, value) then table.insert(self.config.targets, 1, value) end
                            else
                                print("Could not find player '"..tostring(value).."' in your raid or party.")
                            end
                        end,
                    },
                },
            },
            
            clear = {
                type = 'execute',
                name = 'Clear all MAs/MTs.',
                func = function() self:clearAll() end,
                order = 50,
            },
            invert = {
                type = 'execute',
                name = 'Invert  assist order',
                func = function() self:invertTargetOrder() end,
                order = 50.1,
            },
            hide = {
                type = 'execute',
                name = 'Hides the addon and disables updates.  Use /ta show to re-enable.',
                func = function() 
                    self.disabled = true
                end,
                order = 50.2,
            },
            show = {
                type = 'execute',
                name = 'Show the addon and re-enables updates after being hidden.  Also shows the addon if it is being hidden because you are not in a group.',
                func = function() 
                    self.disabled = false
                    if not self:playerInGroup() and self.config.hideOutOfGroup then
                        self.config.hideOutOfGroup = false
                        self:updateRoster()
                    end
                end,
                order = 50.3,
            },
            config = {
                type = 'execute',
                name = 'Open configuration panel.',
                func = function() InterfaceOptionsFrame_OpenToCategory(self.optionsBaseFrame) end,
                order = 51,
            },
        }
    }

end


